<?php
/**
 * PHOENIX MEDIA - Cleanup
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to license that is bundled with
 * this package in the file LICENSE.
 *
 * @category   Phoenix
 * @package	   Phoenix_Cleanup
 * @copyright  Copyright (c) 2013-2019 PHOENIX MEDIA GmbH (http://www.phoenix-media.eu)
 */
namespace Phoenix\Cleanup\Model\Handler;

use Magento\Framework\App\Filesystem\DirectoryList;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\Archive;
use Magento\Framework\Filesystem;
use Magento\Framework\Filesystem\Io\File;
use Phoenix\Cleanup\Api\HandlerInterface;
use Phoenix\Cleanup\Helper\Data as Helper;
use Phoenix\Cleanup\Logger\Logger;
use Phoenix\Cleanup\Model\Config;

class Media extends AbstractFiles implements HandlerInterface
{
    /**
     * Magento media path
     *
     * @var string
     */
    protected $magentoMediaPath;

    /**
     * path where the images are copied to until deletion
     *
     * @var string
     */
    protected $mediaRecycleBinPath;

    /**
     * Magento cache path
     *
     * @var string
     */
    protected $cachePath;

    /**
     * Magento watermark path
     *
     * @var string
     */
    protected $watermarkPath;

    /**
     * Magento placeholder path
     *
     * @var string
     */
    protected $placeholderPath;

    /**
     * list of media files in the media directory
     *
     * @var array
     */
    protected $fileList = [];

    /**
     * list of files with no assignment
     *
     * @var array
     */
    protected $deleteList = [];

    /**
     * contains number of files in folders and sub-folders
     *
     * @var array
     */
    protected $directoryHashMap = [];

    /**
     * @var ResourceConnection
     */
    private $resourceConnection;

    public function __construct(
        ResourceConnection $resourceConnection,
        Config $config,
        Logger $logger,
        Helper $helper,
        Filesystem $filesystem,
        DirectoryList $directoryList,
        Archive $archive,
        File $io
    ) {
        parent::__construct($config, $logger, $helper, $filesystem, $directoryList, $archive, $io);

        $this->resourceConnection = $resourceConnection;
    }

    /**
     * Returns is configuration allowed execution
     *
     * @return bool
     */
    public function isEnabled()
    {
        return $this->config->getCleanupMedia();
    }

    /**
     * Runs cleanup
     *
     * @return $this
     */
    public function cleanup()
    {
        $this->init();
        $this->io->mkdir($this->mediaRecycleBinPath);

        $this->getFilesRecursive($this->magentoMediaPath);

        $this->log('detected ' . count($this->fileList) . ' files in media folder');

        $this->createDeleteList();
        $this->deleteFiles();
        $this->cleanupMediaRecycleBin();

        $this->log('files in media directory: ' . count($this->fileList));
        $this->log('files with no assignment: ' . count($this->deleteList));

        $this->log('checking for empty directories in media folder');
        $this->deleteEmptyFolders();
        $this->log('finished checking for empty directories in media folder');

        //output the savings in bytes
        $bytesFormatted = $this->helper->getBytesFormatted($this->sizeCount, 2);
        $this->logger->info(
            'moved to recycle bin: ' . number_format($this->sizeCount, 0, ',', '.') . ' Bytes (' . $bytesFormatted . ')'
        );

        return $this;
    }

    protected function init()
    {
        parent::init(); // TODO: Change the autogenerated stub
        $this->magentoMediaPath    = $this->directoryList->getPath(DirectoryList::MEDIA) . DIRECTORY_SEPARATOR . 'catalog' . DIRECTORY_SEPARATOR . 'product';
        $this->mediaRecycleBinPath = $this->directoryList->getPath(DirectoryList::VAR_DIR) . DIRECTORY_SEPARATOR . 'media_recyclebin';
        $this->cachePath           = $this->magentoMediaPath . DIRECTORY_SEPARATOR . 'cache';
        $this->placeholderPath     = $this->magentoMediaPath . DIRECTORY_SEPARATOR . 'placeholder';
        $this->watermarkPath       = $this->magentoMediaPath . DIRECTORY_SEPARATOR . 'watermark';
    }

    /**
     * remove the files from the recycle bin
     */
    public function cleanupMediaRecycleBin()
    {
        $mediaCleanupDays = $this->config->getKeepMediaDays();
        $this->cleanupArchive($this->mediaRecycleBinPath, $mediaCleanupDays, true);
    }

    /**
     * recursively detect empty folders
     *
     * @param $path
     */
    protected function detectEmptyFolders($path)
    {
        $directories    = glob($path . '/*', GLOB_ONLYDIR);
        $files          = glob($path . '/*.*');

        $pathParts      = explode('/', str_replace($this->magentoMediaPath, '', $path));

        $depth          = count($pathParts) - 1;
        $fileCount      = count($files);
        $currentPath    = $path;

        for ($level = $depth; $level >= 0; $level --) {
            if (isset($pathParts[$level]) && $pathParts[$level] != '') {
                if (empty($this->directoryHashMap[$level][$currentPath]) == true) {
                    $this->directoryHashMap[$level][$currentPath] = $fileCount;
                } else {
                    $this->directoryHashMap[$level][$currentPath] += $fileCount;
                }

                //set path to one level above
                $currentPath = substr($currentPath, 0, (strlen('/' . $pathParts[$level]) * -1));
            } else {
                if ($level != 0) {
                    $currentPath = substr($currentPath, 0, -1);
                }
            }
        }

        //do recursive call
        if (empty($directories) == false) {
            foreach ($directories as $directory) {
                //ignore placeholder and watermark
                if ($directory != $this->placeholderPath && $directory != $this->watermarkPath) {
                    $this->detectEmptyFolders($directory);
                }
            }
        }
    }

    /**
     * delete empty folders in media directory
     */
    protected function deleteEmptyFolders()
    {
        $this->detectEmptyFolders($this->magentoMediaPath);

        //delete the empty folders
        $depth = !empty($this->directoryHashMap) ? max(array_keys($this->directoryHashMap)) : 0;
        for ($level = $depth; $level >= 0; $level --) {
            if (empty($this->directoryHashMap[$level]) == false) {
                foreach ($this->directoryHashMap[$level] as $directory => $fileCount) {
                    if ($fileCount == 0) {
                        try {
                            if ($this->dryRun == false) {
                                rmdir($directory);
                            }
                            $this->log('delete empty folder: ' . $directory);
                        } catch (Exception $e) {
                            $this->log('error deleting empty folder: ' . $e->getMessage());
                        }
                    }
                }
            }
        }
    }

    /**
     * creates a list of files which are going to be deleted
     */
    protected function createDeleteList()
    {
        $this->log('checking database');

        /* @var ResourceConnection $resource */
        $resource   = $this->resourceConnection;
        $connection = $resource->getConnection('core_read');

        $tblCatalogProductEntityMediaGallery = $this->resourceConnection
            ->getTableName('catalog_product_entity_media_gallery');

        foreach ($this->fileList as $file) {
            $query  = "SELECT value FROM " . $tblCatalogProductEntityMediaGallery . " WHERE value = '"
                . $this->sanitizeFilePathForDbLookup($file) . "' LIMIT 1";

            $result = $connection->fetchOne($query);

            if (empty($result) == true) {
                $this->deleteList[] = $file;
            }
        }

        $this->log('done');
    }

    /**
     * get files from a directory recursively
     *
     * @param $path
     */
    protected function getFilesRecursive($path)
    {
        $directories = glob($path . '/*', GLOB_ONLYDIR);

        $files = glob($path . '/*.*');

        if (empty($files) == false) {
            foreach ($files as $file) {
                //check if it i not a directory, as directories may contain a . as well
                if (is_dir($file) == false) {
                    $this->fileList[] = $file;
                }
            }
        }

        if (empty($directories) == false) {
            foreach ($directories as $directory) {
                //ignore placeholder and watermark
                if ($directory != $this->placeholderPath && $directory != $this->watermarkPath) {
                    $this->getFilesRecursive($directory);
                } else {
                    switch ($directory) {
                        case $this->placeholderPath:
                            $this->log('skipping placeholder path');
                            break;
                        case $this->watermarkPath:
                            $this->log('skipping watermark path');
                            break;
                    }
                }

                $this->pingDb();
            }
        }
    }

    /**
     * delete files from the filesystem and move them to the media recycle bin
     */
    protected function deleteFiles()
    {
        if (count($this->deleteList) > 0) {
            //check and create the recycle bin path
            $currentRecycleBinPath = $this->mediaRecycleBinPath . DIRECTORY_SEPARATOR . date('Ymd_His');
            if (file_exists($currentRecycleBinPath) == false) {
                $this->io->mkdir($currentRecycleBinPath, 0777, true);
            }
        }

        $deleteCounter = 0;
        foreach ($this->deleteList as $file) {
            if (file_exists($file) === true) {
                if ($this->dryRun == false) {
                    try {
                        //get recycle bin path
                        $newFilePath = $currentRecycleBinPath . str_replace($this->magentoMediaPath, '', $file);
                        $this->log('moving: ' . $file . ' to recycle bin: ' . $newFilePath);

                        //check if directory exists
                        $pathToCheck = str_replace(basename($newFilePath), '', $newFilePath);
                        if (file_exists($pathToCheck) == false) {
                            $this->io->mkdir($pathToCheck, 0777, true);
                        }

                        //move to recycle bin
                        copy($file, $newFilePath);

                        //delete the original file
                        $this->calculateSavings($file);
                        $this->log('deleting: ' . $file);
                        unlink($file);
                    } catch (Exception $e) {
                        $this->log('error during file cleanup: ' . $e->getMessage());
                    }
                }

                $deleteCounter ++;
            }

            $this->pingDb();
        }

        $this->log('deleted ' . $deleteCounter . ' obsolete media files');
    }

    /**
     * @param $file
     * @return array|string|string[]
     */
    protected function sanitizeFilePathForDbLookup($file)
    {
        $file = str_replace($this->magentoMediaPath, '', $file);

        if (substr($file, 0, 6) === '/cache') {
            $fileParts = explode('/', $file);
            array_shift($fileParts);
            array_shift($fileParts);
            array_shift($fileParts);

            $file = '/' . implode('/', $fileParts);
        }

        return $file;
    }
}
